package jemu.system.cpc;

import java.applet.Applet;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import jemu.core.Util;
import jemu.core.cpu.Processor;
import jemu.core.device.Computer;
import jemu.core.device.DeviceMapping;
import jemu.ui.Switches;
import jemu.core.device.crtc.Basic6845;
import jemu.core.device.floppy.Drive;
import jemu.core.device.floppy.UPD765A;
import jemu.core.device.io.PPI8255;
import jemu.core.device.memory.Memory;
import jemu.core.device.sound.AY_3_8910;
import jemu.ui.Display;
import jemu.util.diss.Disassembler;
import jemu.util.diss.DissZ80;
import jemu.ui.Autotype;
import jemu.ui.CPCPrinter;
import jemu.ui.JEMU;
import java.io.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.image.*;
/**
* Title:        JavaCPC
* Description:  The Java CPC Emulator
* Copyright:    Copyright (c) 2002-2008
* Company:
* @author:      Markus, Richard (JEMU)
*/

public class CPC extends Computer {
    public int callstart = 0;
    public int[] SCR_HEADER = {
        0x00,0x4A,0x41,0x56,0x41,0x43,0x50,0x43,
        0x20,0x53,0x43,0x52,0x00,0x00,0x00,0x00,
        0x00,0x00,0x02,0x00,0x00,0x00,0xC0,0x00,
        0x00,0x40,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x40,0x00,0x42,0x04,0x00,0x00,0x4A,
        0x41,0x56,0x41,0x43,0x50,0x43,0x20,0x24,
        0x24,0x24,0xFF,0x00,0xFF,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
    };
    
    public int[] PAL_HEADER = {
        0x00,0x11,0x16,0x0D,0x12,0x16,0x20,0x20,
        0x20,0x50,0x41,0x4C,0x00,0x00,0x00,0x00,
        0x00,0x00,0x02,0xEF,0x00,0x09,0x88,0x00,
        0xEF,0x00,0x09,0x88,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0xEF,0x00,0x00,0x8A,0x05,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00  
    };
    
    public int[] PAL_INK = {
        0x54,0x44,0x55,0x5c,0x58,0x5d,0x4c,0x45,
        0x4d,0x56,0x46,0x57,0x5e,0x40,0x5f,0x4e,
        0x47,0x4f,0x52,0x42,0x53,0x5a,0x59,0x5b,
        0x4a,0x43,0x4b,0x54,0x54,0x54,0x54
    };
    
    public int[] SCR_CODE =  // Code to put to #C7D0
    {
        0x3A, 0xD0, 0xD7,               //      LD      A,  (#D7D0)
        0xCD, 0x1C, 0xBD,               //      CALL    #BD1C
        0x21, 0xD1, 0xD7,               //      LD      HL, #D7D1
        0x46,                           //      LD      B,  (HL)
        0x48,                           //      LD      C,  B
        0xCD, 0x38, 0xBC,               //      CALL    #BC38
        0xAF,                           //      XOR     A
        0x21, 0xD1, 0xD7,               //      LD      HL, #D7D1
        0x46,                           // BCL: LD      B,  (HL)
        0x48,                           //      LD      C,  B
        0xF5,                           //      PUSH    AF
        0xE5,                           //      PUSH    HL
        0xCD, 0x32, 0xBC,               //      CALL    #BC32
        0xE1,                           //      POP     HL
        0xF1,                           //      POP     AF
        0x23,                           //      INC     HL
        0x3C,                           //      INC     A
        0xFE, 0x10,                     //      CP      #10
        0x20, 0xF1,                     //      JR      NZ,BCL
        0xC3, 0x18, 0xBB                //      JP      #BB18
    };
    
    public static boolean FDCReset = false;
    public CPCDiscImage dskImage;
    public CPCDiscImage dskImageA;
    public CPCDiscImage dskImageB;
    public CPCDiscImage dskImageC;
    public CPCDiscImage dskImageD;
    
      JFrame dummy = new JFrame();
    
        protected GridBagConstraints gbcConstraints   = null;
    
    String[] Palette = new String[33];
    Color[] Palcols = {
                          /*R G B */
                new Color(0x000000), /* 0*/
                new Color(0x00007F), /* 1*/
                new Color(0x0000FF), /* 2*/
                new Color(0x7F0000), /* 3*/
		new Color(0x7F007F), /* 4*/
                new Color(0x7F00FF), /* 5*/
                new Color(0xFF0000), /* 6*/
                new Color(0xFF007F), /* 7*/
                new Color(0xFF00FF), /* 8*/
                new Color(0x007F00), /* 9*/
                new Color(0x007F7F), /*10*/
                new Color(0x007FFF), /*11*/
                new Color(0x7F7F00), /*12*/
		new Color(0x7F7F7F), /*13*/
                new Color(0x7F7FFF), /*14*/
                new Color(0xFF7F00), /*15*/
                new Color(0xFF7F7F), /*16*/
                new Color(0xFF7FFF), /*17*/
                new Color(0x00FF00), /*18*/
                new Color(0x00FF7F), /*19*/
                new Color(0x00FFFF), /*20*/
                new Color(0x7FFF00), /*21*/
                new Color(0x7FFF7F), /*22*/
                new Color(0x7FFFFF), /*23*/
                new Color(0xFFFF00), /*24*/
                new Color(0xFFFF7F), /*25*/
                new Color(0xFFFFFF), /*26*/
                new Color(0x7F7F7F), /*27*/
		new Color(0xFF007F), /*28*/
                new Color(0xFFFF7F), /*29*/
		new Color(0x00007F), /*30*/
                new Color(0x00FF7F)  /*31*/};
    
    public JButton savesna = new JButton("");

  public int joyreader;
  protected boolean turbo = false;
  public int autotyper = 0;
  public int readkey  = 0;
  public static boolean blockKeyboard;
  public static int getfromautotype = 0;

  int[] eventArray;
  boolean[] shifter;
  public static boolean shift;

  // Conversion of CPC BDIR & BC1 values 0..3 to PSG values
  protected static final int[] PSG_VALUES = new int[] {
    AY_3_8910.BC2_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BC1_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BDIR_MASK,
    AY_3_8910.BC2_MASK | AY_3_8910.BDIR_MASK | AY_3_8910.BC1_MASK
  };

  // Port mappings
  protected static final int   PSG_PORT_A        = -1;
  protected static final int   PPI_PORT_B        = -2;
  protected static final int   PPI_PORT_C        = -3;
  
  public static int autotype = 4;

  protected static final int   CYCLES_PER_SECOND = 1000000;
  protected static final int   AUDIO_TEST        = 0x40000000;

  protected String             lowerROM          = null;
  protected Hashtable          upperROMs         = new Hashtable();
  protected Z80                z80               = new Z80(CYCLES_PER_SECOND);
  protected CPCMemory          memory            = null;                                                      // new
  // CPCMemory(CPCMemory.TYPE_512K);
  protected Basic6845          crtc              = (Basic6845) addDevice(new Basic6845());
  protected GateArray          gateArray         = (GateArray) addDevice(new GateArray(this));
  protected PPI8255            ppi               = (PPI8255) addDevice(new PPI8255());
  protected AY_3_8910          psg               = (AY_3_8910) addDevice(new AY_3_8910());
  protected UPD765A            fdc               = (UPD765A) addDevice(new UPD765A(4));  // 4 cycles per call to cycle()
  protected Disassembler       disassembler      = new DissZ80();
  protected int                audioAdd          = psg.getSoundPlayer().getClockAdder(AUDIO_TEST,
                                                                                           CYCLES_PER_SECOND);
  protected int                audioCount        = 0;
  protected Drive[]            floppies          = new Drive[4];
  protected int                currentDrive      = 0;

  protected KeyboardB          keyboardb         = new KeyboardB();
  protected KeyboardA          keyboarda         = new KeyboardA();
  protected KeyboardN          keyboardn         = new KeyboardN();

  public CPC(Applet applet, String name) {
    super(applet, name);
    initCPCType(name);
    z80.setMemoryDevice(memory);
    z80.addOutputDeviceMapping(new DeviceMapping(memory, 0x2000, 0x0000)); // ROM Select
    z80.setInterruptDevice(gateArray);
    z80.addOutputDeviceMapping(new DeviceMapping(gateArray, 0xc000, 0x4000)); // All GA functions
    z80.setCycleDevice(this);
    crtc.setRegisterSelectMask(0x0100, 0x0000);
    crtc.setCRTCListener(gateArray);
    z80.addOutputDeviceMapping(new DeviceMapping(crtc, 0x4000, 0x0000));
    ppi.setPortMasks(0x0100, 0x0100, 0x0200, 0x0200);
    ppi.setReadDevice(PPI8255.PORT_B, this, PPI_PORT_B);
    ppi.setWriteDevice(PPI8255.PORT_C, this, PPI_PORT_C);
    ppi.setReadDevice(PPI8255.PORT_A, psg, 0);
    ppi.setWriteDevice(PPI8255.PORT_A, psg, 0);
    psg.setReadDevice(AY_3_8910.PORT_A, this, PSG_PORT_A);
    psg.setClockSpeed(CYCLES_PER_SECOND);
    z80.addOutputDeviceMapping(new DeviceMapping(ppi, 0x0800, 0x0000));
    z80.addInputDeviceMapping(new DeviceMapping(ppi, 0x0800, 0x0000));
    z80.addOutputDeviceMapping(new DeviceMapping(fdc, 0x0580, 0x0100));
    z80.addInputDeviceMapping(new DeviceMapping(fdc, 0x0580, 0x0100));
    z80.addOutputDeviceMapping(new DeviceMapping(this, 0x581, 0x0000));
    for (int i = 0; i < 4; i++)
      fdc.setDrive(i, floppies[i] = new Drive(i == 0 ? 1 : 2));
    setBasePath("cpc");
  }
  
  private void initCPCType(String name) {
            
    // check for MAXAM assembler
    if (Switches.Maxam)
      upperROMs.put("14", "MAXAM.zip");    
    
    // Set memorytype for all models
      memory = new CPCMemory(CPCMemory.TYPE_512K);
      if (Switches.Memory.equals("TYPE_512K"))
        memory = new CPCMemory(CPCMemory.TYPE_512K);
      if (Switches.Memory.equals("TYPE_256K"))
        memory = new CPCMemory(CPCMemory.TYPE_256K);
      if (Switches.Memory.equals("TYPE_128K"))
        memory = new CPCMemory(CPCMemory.TYPE_128K);
      if (Switches.Memory.equals("TYPE_64K"))
        memory = new CPCMemory(CPCMemory.TYPE_64K);
      if (Switches.Memory.equals("TYPE_SILICON_DISC"))
        memory = new CPCMemory(CPCMemory.TYPE_SILICON_DISC);
    //
      
      if ("CPC464".equals(name)) {
      lowerROM = "OS464.zip";
      upperROMs.put("0", "BASIC1-0.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC664".equals(name)) {
      lowerROM = "OS664.zip";
      upperROMs.put("0", "BASIC664.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC464PARA".equals(name)) {
      lowerROM = "OS464.zip";
      upperROMs.put("0", "BASIC1-0.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("CPC6128".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("ST128".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "STK1.zip");
      upperROMs.put("2", "STK2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("ART".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "ART0.zip");
      upperROMs.put("2", "ART1.zip");
      upperROMs.put("3", "ART2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("MEGA".equals(name)) {
      lowerROM = "MOS6128.ZIP";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "STK1.zip");
      upperROMs.put("2", "STK2.zip");
      upperROMs.put("3", "ART0.zip");
      upperROMs.put("4", "ART1.zip");
      upperROMs.put("5", "ART2.zip");
      upperROMs.put("6", "locksmit.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("8", "sym-romA.zip");
      upperROMs.put("9", "sym-romB.zip");
      upperROMs.put("10", "sym-romC.zip");
      upperROMs.put("11", "sym-romD.zip");
      upperROMs.put("12", "ROMPKP1.zip");
      upperROMs.put("13", "ROMPKP2.zip");
    } else if ("PARADOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("ROMPACK".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "ROMPKP1.zip");
      upperROMs.put("2", "ROMPKP2.zip");
      upperROMs.put("7", "PARADOS.zip");
    } else if ("ROMSYM".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("1", "sym-romA.zip");
      upperROMs.put("2", "sym-romB.zip");
      upperROMs.put("3", "sym-romC.zip");
      upperROMs.put("4", "sym-romD.zip");
      upperROMs.put("5", "ROMPKP1.zip");
      upperROMs.put("6", "ROMPKP2.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC6128fr".equals(name)) {
      lowerROM = "OSFR.zip";
      upperROMs.put("0", "BASICFR.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPC6128es".equals(name)) {
      lowerROM = "OSES.zip";
      upperROMs.put("0", "BASICES.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("SymbOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("1", "sym-romA.zip");
      upperROMs.put("2", "sym-romB.zip");
      upperROMs.put("3", "sym-romC.zip");
      upperROMs.put("4", "sym-romD.zip");
      } else if ("FutureOS".equals(name)) {
      lowerROM = "OS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
      upperROMs.put("10", "T-OG-E-A.zip");
      upperROMs.put("11", "T-OG-E-B.zip");
      upperROMs.put("12", "T-OG-E-C.zip");
      upperROMs.put("13", "T-OG-E-D.zip");
    } else if ("KCcomp".equals(name)) {
      lowerROM = "KCCOS.zip";
      upperROMs.put("0", "KCCBAS.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("COMCPC".equals(name)) {
      lowerROM = "-OS.zip";
      upperROMs.put("0", "-BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("JEMCPC".equals(name)) {
      lowerROM = "jemuOS6128.zip";
      upperROMs.put("0", "BASIC1-1.zip");
      upperROMs.put("7", "AMSDOS.zip");
    } else if ("CPCPLUS".equals(name)) {
      lowerROM = "os-plus.zip";
      upperROMs.put("0", "basic-plus.zip");
      upperROMs.put("7", "romplus.zip");
    } else if ("RAMSETTER".equals(name)) {
      lowerROM = "OS6128.zip";
    }
    // Manual rom settings
    // in progress
      
    
    else if ("CUSTOM".equals(name)) {
      if (!Switches.LOWER.equals("none"))    lowerROM = Switches.LOWER;
      if (!Switches.UPPER_0.equals("none"))  upperROMs.put("0", Switches.UPPER_0);
      if (!Switches.UPPER_1.equals("none"))  upperROMs.put("1", Switches.UPPER_1);
      if (!Switches.UPPER_2.equals("none"))  upperROMs.put("2", Switches.UPPER_2);
      if (!Switches.UPPER_3.equals("none"))  upperROMs.put("3", Switches.UPPER_3);
      if (!Switches.UPPER_4.equals("none"))  upperROMs.put("4", Switches.UPPER_4);
      if (!Switches.UPPER_5.equals("none"))  upperROMs.put("5", Switches.UPPER_5);
      if (!Switches.UPPER_6.equals("none"))  upperROMs.put("6", Switches.UPPER_6);
      if (!Switches.UPPER_7.equals("none"))  upperROMs.put("7", Switches.UPPER_7);
      if (!Switches.UPPER_8.equals("none"))  upperROMs.put("8", Switches.UPPER_8);
      if (!Switches.UPPER_9.equals("none"))  upperROMs.put("9", Switches.UPPER_9);
      if (!Switches.UPPER_A.equals("none"))  upperROMs.put("10", Switches.UPPER_A);
      if (!Switches.UPPER_B.equals("none"))  upperROMs.put("11", Switches.UPPER_B);
      if (!Switches.UPPER_C.equals("none"))  upperROMs.put("12", Switches.UPPER_C);
      if (!Switches.UPPER_D.equals("none"))  upperROMs.put("13", Switches.UPPER_D);
      if (!Switches.UPPER_E.equals("none"))  upperROMs.put("14", Switches.UPPER_E);
      if (!Switches.UPPER_F.equals("none"))  upperROMs.put("15", Switches.UPPER_F);
    }
    Switches.computersys = 1;
    if (Switches.Expansion)        
      upperROMs.put("6", "JAVACPC.zip");
  }

  
  public void initialise() {

    String romNum;
    String romName;
    memory.setLowerROM(getFile(romPath + lowerROM));
    Set roms = upperROMs.keySet();
    Iterator romIt = roms.iterator();
    while (romIt.hasNext()) {
      romNum = (String) romIt.next();
      romName = (String) upperROMs.get(romNum);
      memory.setUpperROM((new Byte(romNum)).byteValue(), getFile(romPath + romName));
    }
    upperROMs.clear();
    gateArray.setMemory(memory.getMemory());
    super.initialise();
    psg.getSoundPlayer().play();
    fdc.initialise();
    fdc.initialise();
    for (int i = 0; i < 4; i++)
      fdc.setDrive(i, null);
  }

  
  public void reset() {
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          
   Display.autotype=0;
      try {
          keyboarda.keyReleased(eventArray[readkey]);
      }
      catch (Exception e) {}
    blockKeyboard = false;
    autotyper = 0;
    super.reset();
    fdc.reset();
    memory.reset();
    psg.getSoundPlayer().resync();
    fdc.initialise();
  }

  
  public void eject() {
    fdc.setDrive(getCurrentDrive(), null);
    int drive = getCurrentDrive();
    if (drive == 0)
        dskImageA = null;
    if (drive == 1)
        dskImageB = null;
    if (drive == 2)
        dskImageC = null;
    if (drive == 3)
        dskImageD = null;
    if (Switches.floppysound)
        UPD765A.ejectSound.play();
    {
      if (getCurrentDrive() == 0)
          Switches.loaddrivea = "Drive is empty.";
      if (getCurrentDrive() == 1)
          Switches.loaddriveb = "Drive is empty.";
      if (getCurrentDrive() == 2)
          Switches.loaddrivec = "Drive is empty.";
      if (getCurrentDrive() == 3)
          Switches.loaddrived = "Drive is empty.";
    }
  }

  
  public void dispose() {
    super.dispose();
    psg.getSoundPlayer().dispose();
  }
  
  public void cycle() {
      if (FDCReset){
          fdc.resetb();
          FDCReset = false;
      }
    fdc.cycle();
    gateArray.cycle();
    if ((audioCount += audioAdd / Switches.turbo) >= AUDIO_TEST) {
      psg.writeAudio();
      audioCount -= AUDIO_TEST;
    }
  }

  public void setDisplay(Display value) {
    super.setDisplay(value);
    gateArray.setDisplay(value);
  }

  public void setFrameSkip(int value) {
    super.setFrameSkip(value);
    gateArray.setRendering(value == 0);
  }

  public void vSync() {
      if (Switches.scores) {
          Switches.scores = false;
          getScores();
      }
      if (Switches.poke){
          Switches.poke = false;
          POKE();
      }
      
      if (Switches.saveScr){
          Switches.saveScr = false;
          exportScreen();
      }
          
      if (Switches.showPalette){
          Switches.showPalette = false;
          showPalette();
      }
      
    if (frameSkip == 0)
      display.updateImage(true);
    syncProcessor(psg.getSoundPlayer());
    typeAuto();
    joyReader();
    
      if (Switches.dskcheck == true){
          Switches.dskcheck = false;
          SaveDSK();
      }
      if (Switches.save64){
          Switches.save64=false;
          SNA_Save(64);
      }
      if (Switches.save128){
          Switches.save128=false;
          SNA_Save(128);
      }
      if (Switches.save256){
          Switches.save256=false;
          SNA_Save(256);
      }
      if (Switches.save512){
          Switches.save512=false;
          SNA_Save(512);
      }
    if (Switches.BINImport){
        Switches.BINImport = false;
        BIN_Load();
    }
  }
  
  public void SaveDSK(){
      dskImage = null;
      int drive = getCurrentDrive();
      System.out.println("Saving image from drive DF" + drive);
      if (drive == 0){
          if (dskImageA != null)
                dskImage=dskImageA;
      }
      if (drive == 1){
          if (dskImageB != null)
                dskImage=dskImageB;
      }
      if (drive == 2){
          if (dskImageC != null)
                dskImage=dskImageC;
      }
      if (drive == 3){
          if (dskImageD != null)
                dskImage=dskImageD;
      }
      if (dskImage != null){
        floppies[drive].setSides(dskImage.getNumberOfSides());
        int heads = dskImage.getNumberOfSides() == 1 ? Drive.HEAD_0 : Drive.BOTH_HEADS;
        fdc.setDrive(drive, floppies[drive]);
        fdc.getDrive(drive).setDisc(heads, dskImage);
        dskImage.saveImage();
      } 
      else
          System.out.println("Failed to save... Probably drive empty?");
      setCurrentDrive(0);
  }
  
  public void joyReader(){
       if (!blockKeyboard){
      if (Switches.MouseJoy){
      if (Switches.directxR.equals("Right"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD6);
      if (Switches.directxL.equals("Left"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD4);
      if (Switches.directyU.equals("Up"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD8);
      if (Switches.directyD.equals("Down"))
          keyboarda.keyPressed(KeyEvent.VK_NUMPAD2);
              Switches.directL++;
              if (Switches.directL >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD4);
              }
              Switches.directR++;
              if (Switches.directR >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD6);
              }
              Switches.directU++;
              if (Switches.directU >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD8);
              }
              Switches.directD++;
              if (Switches.directD >= 4){
                  keyboarda.keyReleased(KeyEvent.VK_NUMPAD2);
              }
      }    
       }
  }

  public void typeAuto(){
    if (getfromautotype != 0) {
      getfromautotype = 0;
      AutoTyper(Autotype.autotext);
    }
    if (autotyper != 0) {
        Switches.turbo=3;
      Display.autotype=2;
      autotyper++;
      if (autotyper == 3) {
        try {
         // System.out.println("KeyRelease: " + readkey);
          keyboarda.keyReleased(eventArray[readkey]);
          if (shifter[readkey + 1])
            keyboarda.keyPressed(KeyEvent.VK_SHIFT);
          else
            keyboarda.keyReleased(KeyEvent.VK_SHIFT);
        }  catch (Exception e) {
          keyboarda.keyReleased(KeyEvent.VK_SHIFT);
          autotyper = 0;
          Display.autotype=0;
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          blockKeyboard = false;
        }
      }
        
      if (autotyper == 5) {
        try {
          //System.out.println("KeyPress: " + readkey);
          readkey++;
          keyboarda.keyPressed(eventArray[readkey]);
          autotyper = 1;
        } catch (Exception e) {
          keyboarda.keyReleased(KeyEvent.VK_SHIFT);
          autotyper = 0;
          Display.autotype=0;
          if (turbo)
              Switches.turbo = 3;
          else 
              Switches.turbo = 1;
          blockKeyboard = false;
        } 
      }
    }
  }

  public Memory getMemory() {
    return memory;
  }

  public Processor getProcessor() {
    return z80;
  }

  public Dimension getDisplaySize(boolean large) {
    return gateArray.getDisplaySize(large);
  }
  
  public Dimension getDisplayScale(boolean large) {
    return large ? Display.SCALE_1x2 : Display.SCALE_1;
  }
  
  public void setLarge(boolean value) {
    gateArray.setHalfSize(!value);
  }

  public Disassembler getDisassembler() {
    return disassembler;
  }
  
  protected Keyboard getKeyboard() {
    if (Switches.joystick == 0)
      return keyboardb;
    else if (Switches.notebook)
      return keyboardn;
    else
      return keyboarda;
  }

  public int readPort(int port) {
int result;
    switch(port) {
      case PPI_PORT_B:
          if (Switches.Printer)
        result = 0x1e | (crtc.isVSync() ? 0x01 : 0);
          else              
        result = 0x5e | (crtc.isVSync() ? 0x01 : 0);
              
        break;
        
      case PSG_PORT_A: {
        if (blockKeyboard) {
          result = keyboarda.readSelectedRow();
          break;
        }
        else {
          result = getKeyboard().readSelectedRow();
          break;
        }
      }
      
      default:
        throw new RuntimeException("Unexpected Port Read: " + Util.hex((short)port));
    }
    return result;
  }

  public void writePort(int port, int value) {
      
    switch(port) {

      case PPI_PORT_C:
        psg.setBDIR_BC2_BC1(PSG_VALUES[value >> 6],ppi.readOutput(PPI8255.PORT_A));
      
        if (blockKeyboard)
          keyboarda.setSelectedRow(value & 0x0f);
        else
          getKeyboard().setSelectedRow(value & 0x0f);
        break;

      default:
        if ((port & 0x0581) == 0) { // FDC Motor control
          System.out.println("Floppy-motor " + ((value & 0x01) == 0 ? "off" : "on"));
          if ((value & 0x01) == 0) {
            if (Switches.floppysound)
              UPD765A.motorSound.play();
            JEMU.menue7.setEnabled(false);
            Display.ledOn = 0;
          }
          else {
            if (Switches.floppysound){
              UPD765A.motorSound.loop();
            JEMU.menue7.setEnabled(true);
            Display.ledOn = 1;}
          }
        } else {
          throw new RuntimeException("Unexpected Port Write: " + Util.hex((short) port) + " with "
            + Util.hex((byte) value));
      }
    }
  }

  public void keyPressed(KeyEvent e) {
    if (!blockKeyboard)
      getKeyboard().keyPressed(e.getKeyCode());
  }

  public void keyReleased(KeyEvent e) {
    if (!blockKeyboard)
      getKeyboard().keyReleased(e.getKeyCode());
  }
  
  public void AtKey() {
    AutoTyper("@");
  }
  
  public void bootDisk() {
      Display.bootgames = 150;
    AutoTyper("mode 1:pen 0\n|A:CAT:FOR k=1 TO 8:n$=n$+chr$(peek(HIMEM-&7FF+k)):NEXT:" +
            "n$=n$+\".\":FOR k=1 TO 3:n$=n$+chr$(peek(HIMEM-&7FF+k+8)):NEXT:PEN 1:RUN n$\n");
  }  
  public void bootDiskb() {
      Display.bootgames = 150;
    AutoTyper("mode 1:pen 0\n|B:CAT:FOR k=1 TO 8:n$=n$+chr$(peek(HIMEM-&7FF+k)):NEXT:" +
            "n$=n$+\".\":FOR k=1 TO 3:n$=n$+chr$(peek(HIMEM-&7FF+k+8)):NEXT:PEN 1:RUN n$\n");
  }  
  
  public void bootCPM() {
    AutoTyper("|CPM\n");
  }
  
  public void typeRun() {
    AutoTyper("RUN\"");      
  }
  
  public void typeCat() {
    AutoTyper("CAT\n");      
  }
  
  public void AutoType() {
    AutoTyper(Switches.loadauto);      
  }  
  
  public void MouseFire1(){
       if (!blockKeyboard)
      keyboarda.keyPressed(KeyEvent.VK_NUMPAD5);
  }
  public void MouseFire2(){
       if (!blockKeyboard)
      keyboarda.keyPressed(KeyEvent.VK_NUMPAD0);
  }
  public void MouseReleaseFire1(){
       if (!blockKeyboard)
      keyboarda.keyReleased(KeyEvent.VK_NUMPAD5);
  }
  public void MouseReleaseFire2(){
       if (!blockKeyboard)
      keyboarda.keyReleased(KeyEvent.VK_NUMPAD0);
      
  }
  public void AutoTyper(String cmd) {
          if (Switches.turbo == 3)
        turbo = true;
    else
        turbo = false;
    blockKeyboard = true;
    readkey = 0;
    eventArray = new int[cmd.length()];
    shifter = new boolean[cmd.length()];
    for(int i = 0; i < cmd.length(); i++) {
      eventArray[i] = convertCharToVK(cmd.charAt(i));
      shifter[i] = shift;
    }
    if (shifter[0])
      keyboarda.keyPressed(KeyEvent.VK_SHIFT);
    keyboarda.keyPressed(eventArray[0]);
    autotyper = 1;
  }
  
  /**
   * This should really be done using CPC key codes and overriding the value returned in
   * readPort for the PSG port read.
   */
  protected static final int[] ASC_TO_KEY = {
    KeyEvent.VK_SPACE, KeyEvent.VK_1, KeyEvent.VK_2, KeyEvent.VK_3,                       // SPACE !"#   FTTT
    KeyEvent.VK_4, KeyEvent.VK_5, KeyEvent.VK_6, KeyEvent.VK_7,                           // $%&'        TTTT
    KeyEvent.VK_8, KeyEvent.VK_9, KeyEvent.VK_SEMICOLON, KeyEvent.VK_QUOTE,               // ()*+        TTTT
    KeyEvent.VK_COMMA, KeyEvent.VK_MINUS, KeyEvent.VK_PERIOD, KeyEvent.VK_SLASH,          // ,-./        FFFF
    KeyEvent.VK_0, KeyEvent.VK_1, KeyEvent.VK_2, KeyEvent.VK_3,                           // 0123        FFFF
    KeyEvent.VK_4, KeyEvent.VK_5, KeyEvent.VK_6, KeyEvent.VK_7,                           // 4567        FFFF
    KeyEvent.VK_8, KeyEvent.VK_9, KeyEvent.VK_SEMICOLON, KeyEvent.VK_QUOTE,               // 89:;        FFFF
    KeyEvent.VK_COMMA, KeyEvent.VK_MINUS, KeyEvent.VK_PERIOD, KeyEvent.VK_SLASH,          // <=>?        TTTT
    KeyEvent.VK_OPEN_BRACKET, KeyEvent.VK_A, KeyEvent.VK_B, KeyEvent.VK_C,                // @ABC        FTTT
    KeyEvent.VK_D, KeyEvent.VK_E, KeyEvent.VK_F, KeyEvent.VK_G,                           // DEFG        TTTT
    KeyEvent.VK_H, KeyEvent.VK_I, KeyEvent.VK_J, KeyEvent.VK_K,                           // HIJK        TTTT
    KeyEvent.VK_L, KeyEvent.VK_M, KeyEvent.VK_N, KeyEvent.VK_O,                           // LMNO        TTTT
    KeyEvent.VK_P, KeyEvent.VK_Q, KeyEvent.VK_R, KeyEvent.VK_S,                           // PQRS        TTTT
    KeyEvent.VK_T, KeyEvent.VK_U, KeyEvent.VK_V, KeyEvent.VK_W,                           // TUVW        TTTT
    KeyEvent.VK_X, KeyEvent.VK_Y, KeyEvent.VK_Z, KeyEvent.VK_ALT_GRAPH,                   // XYZ[        TTTF
    KeyEvent.VK_BACK_SLASH, KeyEvent.VK_CLOSE_BRACKET, KeyEvent.VK_EQUALS, KeyEvent.VK_0, // \]^_        FFFT
    KeyEvent.VK_BACK_SLASH, KeyEvent.VK_A, KeyEvent.VK_B, KeyEvent.VK_C,                  // `abc        TFFF
    KeyEvent.VK_D, KeyEvent.VK_E, KeyEvent.VK_F, KeyEvent.VK_G,                           // defg        FFFF
    KeyEvent.VK_H, KeyEvent.VK_I, KeyEvent.VK_J, KeyEvent.VK_K,                           // hijk        FFFF
    KeyEvent.VK_L, KeyEvent.VK_M, KeyEvent.VK_N, KeyEvent.VK_O,                           // lmno        FFFF
    KeyEvent.VK_P, KeyEvent.VK_Q, KeyEvent.VK_R, KeyEvent.VK_S,                           // pqrs        FFFF
    KeyEvent.VK_T, KeyEvent.VK_U, KeyEvent.VK_V, KeyEvent.VK_W,                           // tuvw        FFFF
    KeyEvent.VK_X, KeyEvent.VK_Y, KeyEvent.VK_Z, KeyEvent.VK_ALT_GRAPH,                   // xyz{        FFFT
    KeyEvent.VK_OPEN_BRACKET, KeyEvent.VK_CLOSE_BRACKET                                   // |}          TT
  };
  
  protected static final String ASC_TO_SHIFT =
    "FTTTTTTTTTTTFFFFFFFFFFFFFFFFTTTTFTTTTTTTTTTTTTTTTTTTTTTTTTTFFFFTTFFFFFFFFFFFFFFFFFFFFFFFFFFTTT";
  
  public static int convertCharToVK(char in) {
    int result;
    switch(in) {
      case '\n': shift = false; result = KeyEvent.VK_ENTER; break;
      case '£':  shift = true; result = KeyEvent.VK_EQUALS; break;
      case 'ö': shift = false; result = KeyEvent.VK_NUMPAD4; break;
      case 'ä': shift = false; result = KeyEvent.VK_NUMPAD6; break;
      case 'ü': shift = false; result = KeyEvent.VK_NUMPAD8; break;
      case 'ß': shift = false; result = KeyEvent.VK_NUMPAD2; break;
      case 'Ö': shift = false; result = KeyEvent.VK_NUMPAD5; break;
      case 'Ä': shift = false; result = KeyEvent.VK_NUMPAD0; break;
    default:
      if (in >= ' ' && in <= '}') {
        in -= 32;
        result = ASC_TO_KEY[in];
        shift = ASC_TO_SHIFT.charAt(in) == 'T';
      }
      else {
        shift = false;
        result = KeyEvent.CHAR_UNDEFINED;
      }
  }
  return result;
}

  protected static final String SNA_HEADER                             = "MV - SNA";
  protected static       String SNA_EYECATCHER                         = "JavaCPC Snapshot";
  protected static final String k64                                    = "64k Snapshot    "; 
  protected static final String k128                                   = "128k Snapshot   "; 
  protected static final String k256                                   = "256k Snapshot   "; 
  protected static final String k512                                   = "512k Snapshot   "; 
  protected static final String DSK_HEADER                             = "MV - CPC";
  protected static final String DSK_HEADER_EXT                         = "EXTENDED";
  protected static final String CDT_HEADER                             = "ZXTAPE";

  protected static final int    CRTC_FLAG_VSYNC_ACTIVE                 = 0x01;
  protected static final int    CRTC_FLAG_HSYNC_ACTIVE                 = 0x02;
  protected static final int    CRTC_FLAG_HDISP_ACTIVE                 = 0x04;
  protected static final int    CRTC_FLAG_VDISP_ACTIVE                 = 0x08;
  protected static final int    CRTC_FLAG_HTOT_REACHED                 = 0x10;
  protected static final int    CRTC_FLAG_VTOT_REACHED                 = 0x20;
  protected static final int    CRTC_FLAG_MAXIMUM_RASTER_COUNT_REACHED = 0x40;

  protected static final int    SNAPSHOT_ID                            = 0x0000;
  protected static final int    VERSION                                = 0x0010;

  protected static final int    AF                                     = 0x0011;
  protected static final int    BC                                     = 0x0013;
  protected static final int    DE                                     = 0x0015;
  protected static final int    HL                                     = 0x0017;
  protected static final int    R                                      = 0x0019;
  protected static final int    I                                      = 0x001a;
  protected static final int    IFF1                                   = 0x001b;
  protected static final int    IFF2                                   = 0x001c;
  protected static final int    IX                                     = 0x001d;
  protected static final int    IY                                     = 0x001f;
  protected static final int    SP                                     = 0x0021;
  protected static final int    PC                                     = 0x0023;
  protected static final int    IM                                     = 0x0025;
  protected static final int    AF1                                    = 0x0026;
  protected static final int    BC1                                    = 0x0028;
  protected static final int    DE1                                    = 0x002a;
  protected static final int    HL1                                    = 0x002c;

  protected static final int    GA_PEN                                 = 0x002e;
  protected static final int    GA_INKS                                = 0x002f;
  protected static final int    GA_ROM                                 = 0x0040;
  protected static final int    GA_RAM                                 = 0x0041;

  protected static final int    CRTC_REG                               = 0x0042;
  protected static final int    CRTC_REGS                              = 0x0043;

  protected static final int    UPPER_ROM                              = 0x0055;

  protected static final int    PPI_A                                  = 0x0056;
  protected static final int    PPI_B                                  = 0x0057;
  protected static final int    PPI_C                                  = 0x0058;
  protected static final int    PPI_CONTROL                            = 0x0059;

  protected static final int    PSG_REG                                = 0x005a;
  protected static final int    PSG_REGS                               = 0x005b;

  protected static final int    MEM_SIZE                               = 0x006b;

  protected static final int    CPC_TYPE                               = 0x006d;

/*  protected static final int    VER_INT_BLOCK                          = 0x006d;
  protected static final int    VER_MODES                              = 0x006e;*/

  protected static final int    HEADER_SIZE                            = 0x0100;

  public void loadFile(int type, String name) throws Exception {
      Switches.name = name;
    byte[] data = getFile(name);
    if (SNA_HEADER.equals(new String(data, 0, SNA_HEADER.length()).toUpperCase())){
        SNA_Load(name);
        System.out.println("Loading MV - SNA snapshot file...");
    }
    else
    if (DSK_HEADER.equals(new String(data, 0, DSK_HEADER.length()).toUpperCase())){
        DSK_Load(name);
        System.out.println("Loading MV - CPCEMU DSK file...");
    }
    else
    if (DSK_HEADER_EXT.equals(new String(data, 0, DSK_HEADER_EXT.length()).toUpperCase())){
        DSK_Load(name);
        System.out.println("Loading EXTENDED CPC DSK file...");
    }
    else
    if (CDT_HEADER.equals(new String(data, 0, CDT_HEADER.length()).toUpperCase())){
        CDT_Load(name);
        System.out.println("Loading ZXTAPE file...");
    }
    else
        BIN_Load(name);
       // System.out.println("Unrecognized file format");
    psg.getSoundPlayer().resync();
    fdc.initialise();
    }
  
  public void CDT_Load(String name) throws Exception {
    byte[] data = getFile(name);
      Switches.booter = 0;
      //CPCTapeImage image = new CPCTapeImage(name, data);
      //
      // TODO: CDT support
      //
       System.out.println("No CDT/TZX support now...");
  }
  
  public void DSK_Load(String name) throws Exception {
    byte[] data = getFile(name);
      Switches.booter = 0;
      CPCDiscImage image = new CPCDiscImage(name, data);
      int drive = getCurrentDrive();
      if (drive == 0)          
      dskImageA = image;
      if (drive == 1)          
      dskImageB = image;
      if (drive == 2)          
      dskImageC = image;
      if (drive == 3)          
      dskImageD = image;
      floppies[drive].setSides(image.getNumberOfSides());
      int heads = image.getNumberOfSides() == 1 ? Drive.HEAD_0 : Drive.BOTH_HEADS;
      fdc.setDrive(drive, floppies[drive]);
      fdc.getDrive(drive).setDisc(heads, image);
      if (getCurrentDrive() == 0)
        Switches.loaddrivea = name;
      if (getCurrentDrive() == 1)
        Switches.loaddriveb = name;
      if (getCurrentDrive() == 2)
        Switches.loaddrivec = name;
      if (getCurrentDrive() == 3)
        Switches.loaddrived = name;
      
      if (Switches.floppysound){
        UPD765A.insertSound.play();
        UPD765A.motorSound.play();
      }
  }
  
  public void SNA_Load(String name){
    byte[] data = getFile(name);
      Switches.booter = 1;
      z80.setAF(getWord(data, AF));
      z80.setBC(getWord(data, BC));
      z80.setDE(getWord(data, DE));
      z80.setHL(getWord(data, HL));
      z80.setR(data[R]);
      z80.setI(data[I]);
      z80.setIFF1(data[IFF1] != 0);
      z80.setIFF2(data[IFF2] != 0);
      z80.setIX(getWord(data, IX));
      z80.setIY(getWord(data, IY));
      z80.setSP(getWord(data, SP));
      z80.setPC(getWord(data, PC));
      z80.setIM(data[IM]);
      z80.setAF1(getWord(data, AF1));
      z80.setBC1(getWord(data, BC1));
      z80.setDE1(getWord(data, DE1));
      z80.setHL1(getWord(data, HL1));

      gateArray.setSelectedInk(data[GA_PEN]);
      System.out.print("INK READ: ");
      for (int i = 0; i < 0x11; i++){
        gateArray.setInk(i, data[GA_INKS + i]);
        System.out.print(data[GA_INKS + i]);
        if (i>=0x10)
            System.out.println("");
        else
            System.out.print(",");
      }
      
      gateArray.setModeAndROMEnable(memory, data[GA_ROM]);
      memory.setRAMBank(data[GA_RAM]);
      
      crtc.setSelectedRegister(data[CRTC_REG]);
      for (int i = 0; i < 18; i++)
        crtc.setRegister(i, data[CRTC_REGS + i]);

      ppi.setControl(data[PPI_CONTROL] & 0xff | 0x80);
      ppi.setOutputValue(PPI8255.PORT_A, data[PPI_A] & 0xff);
      ppi.setOutputValue(PPI8255.PORT_B, data[PPI_B] & 0xff);
      int portC = data[PPI_C] & 0xff;
      ppi.setOutputValue(PPI8255.PORT_C, portC);

      psg.setBDIR_BC2_BC1(PSG_VALUES[portC >> 6], ppi.readOutput(PPI8255.PORT_A));

      psg.setSelectedRegister(data[PSG_REG]);
      for (int i = 0; i < 14; i++)
        psg.setRegister(i, data[PSG_REGS + i] & 0xff);

      int memSize = (data[MEM_SIZE] & 0xff) * 1024;
      memSize = Math.min(memSize, memory.getRAMType() == CPCMemory.TYPE_512K ? 0x100000 : 0x200000);
      int length = data.length;
      System.out.println("Snapshot length is "+ length);
      if (length <= 65800)
          memSize = 0x10000;
      else if (length <= 131400)
          memSize = 0x20000;
      else if (length <= 262800)
          memSize = 0x40000;
      else if (length <= 524600)
          memSize = 0x80000;
      System.out.println("Calculated length is "+ memSize);
          
      byte[] mem = memory.getMemory();
      System.arraycopy(data, 0x100, mem, 0, memSize);
  }
  
  public void SNA_Save(int sizemem){  
      FileDialog filedia = new FileDialog((Frame) dummy, "Save "+sizemem +"k Snapshot File", FileDialog.SAVE);
        filedia.setFile("*.sna");
        filedia.setVisible(true);
        String filename = filedia.getFile();
        if (filename != null) { 
           filename =  filedia.getDirectory() + filedia.getFile();
           String savename = filename;
            if (!savename.toLowerCase().endsWith(".sna"))
                        savename=savename + ".sna";
        File file = new File(savename);
        
        // save data to file
    try {
      int memSize = 0x10000;
      if (sizemem == 128)
         memSize = 0x20000;
      if (sizemem == 256)
         memSize = 0x40000;
      if (sizemem == 512)
         memSize = 0x80000;

    byte[] data = new byte[0x0100+memSize];
    try {
      System.arraycopy(SNA_HEADER.getBytes("UTF-8"), 0, data, 0, SNA_HEADER.length());
    }
        catch (final IOException iox) {}
    String snasize = k64;
    if (sizemem == 64)
        snasize = k64;
    if (sizemem == 128)
        snasize = k128;
    if (sizemem == 256)
        snasize = k256;
    if (sizemem == 512)
        snasize = k512;
        
    try {
        SNA_EYECATCHER = SNA_EYECATCHER + snasize;
      System.arraycopy(SNA_EYECATCHER.getBytes("UTF-8"), 0, data, 0xe0, SNA_EYECATCHER.length());
    }
        catch (final IOException iox) {}
    
      data[VERSION] = (byte)1;
      
      // z80 registers
      putWord(data,AF,z80.getRegisterValue(1));
      putWord(data,BC,z80.getRegisterValue(7));
      putWord(data,DE,z80.getRegisterValue(5));
      putWord(data,HL,z80.getRegisterValue(3));
      data[R] = (byte)z80.getRegisterValue(14);
      data[I] = (byte)z80.getRegisterValue(12);
      data[IFF1] = (byte)z80.getIFF1();
      data[IFF2] = (byte)z80.getIFF2();
      putWord(data,IX,z80.getRegisterValue(9));
      putWord(data,IY,z80.getRegisterValue(11));
      putWord(data,SP,z80.getRegisterValue(10));
      putWord(data,PC,z80.getRegisterValue(13));
          data[IM] = (byte)z80.getIM();
      putWord(data,AF1,z80.getRegisterValue(2));
      putWord(data,BC1,z80.getRegisterValue(8));
      putWord(data,DE1,z80.getRegisterValue(6));
      putWord(data,HL1,z80.getRegisterValue(4));
      
      
      // inks and pens
      data[GA_PEN] = (byte)gateArray.getSelectedInk();
      System.out.print("CPC Palette is: ");
      for (int i = 0; i < 0x11; i++){
        data[GA_INKS +i] = (byte)gateArray.getInks(i);
          System.out.print((byte)gateArray.getInks(i)+ " ");
      }
      System.out.println();
      // CRTC registers
      data[CRTC_REG] = (byte)crtc.getSelectedRegister();
      for (int i = 0; i < 18; i++)
          data[CRTC_REGS + i] = (byte)crtc.getRegister(i);
      data[PPI_CONTROL] = (byte)ppi.getControl();
        data[PPI_A] = (byte)ppi.portInputA(); 
        data[PPI_B] = (byte)ppi.portInputB(); 
        data[PPI_C] = (byte)ppi.portInputC(); 
      
      data[PSG_REG] = (byte)psg.getSelectedRegister();
      for (int i = 0; i < 14; i++)
      data[PSG_REGS + i] = (byte)psg.getRegister(i);
     
      data[GA_ROM] = (byte)gateArray.getMode();
      data[GA_RAM] = (byte)(gateArray.getRAMBank());

      data[UPPER_ROM] = (byte)memory.getUpperROM();
      data[MEM_SIZE] = (byte)64; 
      if (sizemem == 128)
      data[MEM_SIZE] = (byte)128;
      if (sizemem == 256)
      data[MEM_SIZE] = (byte)256; 
      if (sizemem == 512)
      {
      int SizeInK = 512;
      data[MEM_SIZE] = (byte)(SizeInK & 0x0ff); 
      data[MEM_SIZE+1] = (byte)((SizeInK>>8) & 0x0ff);
      }
      data[CPC_TYPE] = (byte)2;
      byte[] mem = memory.getMemory();
      System.arraycopy(mem,0,data,0x0100, memSize);
        

      final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(file));
            
      for (int i=0; i<0x0100+memSize; i++)
	   	bos.write(data[i]);  
      bos.close();
    } catch (final IOException iox) {
      System.out.println("can't write to file ");
    }
        }
filedia.dispose();
  }
  
  public void BIN_Load(){  
        FileDialog filedia = new FileDialog((Frame) dummy, "Import CPC file...", FileDialog.LOAD);
        filedia.setVisible(true);
        String filename = filedia.getFile();
        if (filename != null) { 
           filename =  filedia.getDirectory() + filedia.getFile();
        BIN_Load(filename);
        }
  }
  
  public void BIN_Load(String name){
      Switches.booter = 1;
      int header = 0;
      int start = 0;
      int execaddress = 0;
      String Sexc="";
      int filetype = 0;
            byte[] data = getFile(name);
            if (CheckAMSDOS(data)){
            int startaddress = getWord(data, 0x15);
                 execaddress = getWord(data , 0x1a);
            int filelength = getWord(data , 0x18);
            filetype = getWord(data,0x12);
            String ftype="";
            if(filetype == 0)
                ftype="BASIC";
            if(filetype == 1)
                ftype="BASIC (protected)";
            if(filetype == 2)
                ftype="BINARY";
            if(filetype == 3)
                ftype="BINARY (protected)";
            if(filetype == 4)
                ftype="IMAGE";
            if(filetype == 5)
                ftype="IMAGE (protected)";
            if(filetype == 6)
                ftype="ASCII";
            if(filetype == 7)
                ftype="ASCII (protected)";
            
             start = startaddress;
             String Sadr = Util.hex(start).substring(4);
             String Slen = Util.hex(filelength).substring(4);
             Sexc = Util.hex(execaddress).substring(4);
             Object[] oheader = {"AMSDOS header found...\n\n" +
                     "&" + Sadr +" - Load address\n"+
                     "&" + Slen +" - File length\n" +
                     "&" + Sexc +" - Exec. address\n\n" +
                     "Filetype is: " + ftype+"\n\n" +
                     "Enter load address:\n(Hexadecimal)"};
             String selectedHeader = JOptionPane.showInputDialog(dummy, oheader,Sadr);
             if (selectedHeader == null) return;
                    if (!selectedHeader.equals(""))
                        try{ 
                            start = Util.hexValue(selectedHeader);
                            if (start != Util.hexValue(Sadr))
                            execaddress = 0;
                            }
                        catch (final Exception iox) {return;}
             if (filetype==0)
                JOptionPane.showMessageDialog(null, "You try to import a BASIC file...\nThe result will not be perfect.\nplease use only on small BASIC files.");
                header = 0x80;
            }
            else {
                   String result = JOptionPane.showInputDialog(dummy, "No AMSDOS header found...\n\n" +
                            "Please enter start address:\n(Hexadecimal)\n");
                    if (result == null) return;
                    if (!result.equals(""))
                        try{ 
                            start = Util.hexValue(result);
                            }
                        catch (final Exception iox) {return;}
                    else
                        return;
                header = 0x00;
            }
            byte[] mem = memory.getMemory();
            int memSize = data.length;
            if ((start + (memSize-header)) >=0x10001)
                JOptionPane.showMessageDialog(null, "An error occured during importing file:\n" + "no or wrong start address entered...\n");

            else
            // 
           /* {
                
                System.arraycopy(data, header, mem, start, memSize-header);
            }*/
                 {
             for (int i = header; i < data.length; i++){ 
                 POKE((start + i - header),data[i]);
             }  
            }
             if (CheckAMSDOS(data)){
                 if (execaddress !=0 || filetype ==0){
            Object[] object = {"Execute program ?"};
            int selectedValue = JOptionPane.showOptionDialog(dummy, object, "Please choose:",
                                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE,null,null,null);
                            if(selectedValue == JOptionPane.YES_OPTION) {
                                if (execaddress !=0 && (filetype >=2 && filetype <=3)){
                                        z80.setBC1(0x07f8c);
                                        z80.setAF1(0x08c80);
                                        z80.setBC((z80.getRegisterValue(7)&0x0ff00) | 0x0ff) ;
                                    CALL(execaddress);
                                    // AutoTyper("CALL &" + Sexc + "\n"); 
                                }
                                    else AutoTyper("RUN\n");
                                   /* {
                                        z80.setBC1(0x07f8c);
                                        z80.setAF1(0x08c80);
                                        z80.setBC((z80.getRegisterValue(7)&0x0ff00) | 0x0ff) ;
                                        z80.setHL(execaddress);
                                        CALL(0xbd13);
                                    }*/
                            }
                 } 
             }
  }
  
  public void CALL(int address){
      z80.setPC(address);
  }
  public void OUT(int address, int value){
      //
  }
  
  public void POKE(){
            int address = 0;
            int value = 0;
            
            JTextField Address = new JTextField();
		JTextField Value = new JTextField();
                Object[] message = {"Address", Address, 
        		"Value", Value};
 
                JOptionPane pane = new JOptionPane( message, 
                                                JOptionPane.WARNING_MESSAGE, 
                                                JOptionPane.OK_CANCEL_OPTION);
                
                pane.createDialog(null, "Poke Memory").setVisible(true);
                
             if (Address.getText().equals("")){                
                 JOptionPane.showMessageDialog(null, "nothing poked...");
                 return;
             }
             String Address1 = Address.getText();
             String Value1 = Value.getText();
             try {
             address = Util.hexValue(Address1);
             value = Util.hexValue(Value1);
             POKE (address, value);
             }
                        catch (final Exception iox) {              
                 JOptionPane.showMessageDialog(null, "nothing poked...");
                 return;
                        }
  }
  
  public void POKE(int address, int value){
      memory.writeByte(address, value);
  }
  
  public void displayLostFocus() {
        keyboardb.reset();
        keyboarda.reset();
        keyboardn.reset();
  }
  
  public Drive[] getFloppyDrives() {
    return floppies;
  }
  
  private GridBagConstraints getGridBagConstraints(int x, int y, double weightx,double weighty,
            int width, int fill)
                {
                if (this.gbcConstraints == null) {
                    this.gbcConstraints = new GridBagConstraints();
                }
                this.gbcConstraints.gridx = x;
                this.gbcConstraints.gridy = y;
                this.gbcConstraints.weightx = weightx;
                this.gbcConstraints.weighty = weighty;
                this.gbcConstraints.gridwidth = width;
                this.gbcConstraints.fill = fill;
            return this.gbcConstraints;
  }
  
  public void showPalette(){
      JFrame Output = new JFrame("CPC Palette");
      Output.setAlwaysOnTop(true);
      Output.getContentPane().setLayout(new GridBagLayout());
      for (int i = 0; i < 17; i++){
          Palette[i] = ""+gateArray.getInk(i);
          String pals;
          pals = Palette[i];
          int cols;
          cols = gateArray.getInk(i);
          String gcols = ""+(byte)gateArray.getInks(i);
          
          if (gateArray.getInk(i) <= 9)
              pals = "0" + pals;
          int x = i;
          int y = 1;
          int w = 1;
          if (i >= 16){
              pals = "BORDER " + pals;
              y = 4;
              x = 0;
              w = 5;
          }
          JButton Palettes = new JButton(pals);
          JButton Paletteg = new JButton(gcols);
          JButton Paletten = new JButton(" ");
          Paletten.setBackground(Palcols[cols]);
          Paletten.setFocusable(false);
          Paletten.setBorder(new BevelBorder(BevelBorder.RAISED)); 
          Palettes.setBackground(Color.DARK_GRAY);
          Palettes.setForeground(Color.LIGHT_GRAY);
          Palettes.setFont(new Font("Courier",1,16));
          Palettes.setBorder(new BevelBorder(BevelBorder.LOWERED)); 
          Palettes.setFocusable(false);
          Paletteg.setBackground(Color.WHITE);
          Paletteg.setForeground(Color.BLACK);
          Paletteg.setFont(new Font("Courier",1,16));
          Paletteg.setBorder(new BevelBorder(BevelBorder.LOWERED)); 
          Paletteg.setFocusable(false);
          
                Output.add(Paletten,getGridBagConstraints(x, y, 1.0, 1.0, w, GridBagConstraints.BOTH));
                Output.add(Palettes,getGridBagConstraints(x, y+1, 1.0, 1.0, w, GridBagConstraints.BOTH));
                Output.add(Paletteg,getGridBagConstraints(x, y+2, 1.0, 1.0, w, GridBagConstraints.BOTH));
      }
      
            BufferedImage images = new BufferedImage(288, 203,  BufferedImage.SCALE_SMOOTH);
            images.getGraphics().drawImage(Switches.image, 0,0, 288, 203,null);
            JLabel Display     =   new JLabel(new ImageIcon(images));
                Output.add(Display,getGridBagConstraints(5, 4, 1.0, 1.0, 12, GridBagConstraints.BOTH));
      Output.pack();
      Output.setResizable(false);
      Output.setVisible(true);
  }
  
  public void exportScreen(){
          FileDialog filedia = new FileDialog((Frame) dummy, "Save 16k CPC Screen File + PAL", FileDialog.SAVE);
        filedia.setFile("*.scr");
        filedia.setVisible(true);
        String filename = filedia.getFile();
        if (filename != null) { 
           filename =  filedia.getDirectory() + filedia.getFile();
           String savename = filename;
           String palname = filename;
           
                   int length = savename.length();
                   String temp = savename;
            if (savename.toLowerCase().endsWith(".scr"))
             temp = savename.substring(0, (length - 4));
            if (savename.toLowerCase().endsWith(".bin"))
             temp = savename.substring(0, (length - 4));
                        savename=temp + ".scr";
                        palname=temp + ".pal";
        File file = new File(savename);
        
        // save data to file
    try {
      int memSize = 0x4000;
        byte[] data = new byte[0x080+memSize];
        for (int i = 0; i < 0x80; i++)
            data[i] = (byte)SCR_HEADER[i]; 
      
      byte[] mem = memory.getMemory();
      System.arraycopy(mem,0xc000,data,0x80, memSize - 1);
      
        
      data[0x17d0 + 0x80] = (byte)gateArray.getScreenMode();
      
      int adr = 0x17d1 + 0x80;
      for (int i = 0; i< 16; i++){
          data[adr+i] = (byte)gateArray.getInk(i);
      }
      adr = 0x07d0 + 0x80;
        for (int i = 0; i < 36; i++){
            data[adr+i] = (byte)SCR_CODE[i];
        }
      boolean RETURN = false;
      Object[] object = { "You can set the screen-mode and inks\n" +
              "with a CALL &C7D0 after you loaded\n" +
              "an exported screen.\n" +
              "Please choose now:\n" +
              "YES:\nWait for a key before return to BASIC", "NO:\nReturn to BASIC after a CALL &C7D0" };
      int selectedValue = JOptionPane.showOptionDialog(dummy, object, "Please choose:",
                                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE,null,null,null);
                            if(selectedValue == JOptionPane.NO_OPTION) {
                                RETURN = true;
                            } 
      if (RETURN){
      data[adr+33] = (byte)0xc9; data[adr+34] = (byte)0x00; data[adr+35] = 0x00;
      }
      
      final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(file));
            
      for (int i=0; i<0x080+memSize; i++)
	   	bos.write(data[i]);  
      bos.close();
      exportPal(palname);
    }
    
        catch (final IOException iox) {}
  }
}
  public void exportPal(String name){
 
        File file = new File(name);
        
        // save data to file
    try {
      int memSize = 0xef;
        byte[] data = new byte[0x080+memSize];
        for (int i = 0; i < 0x80; i++)
            data[i] = (byte)PAL_HEADER[i]; 
        data[0x80] = (byte)gateArray.getScreenMode();
        int adr = 0x83;
        for (int i = 0; i < 17; i++)
            for (int g = 0; g < 12; g++){
                int numb = gateArray.getInk(i);
                data[adr] = (byte)PAL_INK[numb];
                adr++;
            }
      
      final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(file));
            
      for (int i=0; i<0x080+memSize; i++)
	   	bos.write(data[i]);  
      bos.close();
    }
    
        catch (final IOException iox) {}
  }
  
/* calculate checksum as AMSDOS would for the first 66 bytes of a datablock */
/* this is used to determine if a file has a AMSDOS header */
    public int ChecksumAMSDOS(byte pHeader[])
    {
        int Checksum;
        int i;
        Checksum = 0;
        for (i=0; i<67; i++)
        {
                int CheckSumByte;
                CheckSumByte = pHeader[i] & 0x0ff;
                Checksum+=CheckSumByte;
        }
        return Checksum;
    }
    public boolean CheckAMSDOS(byte pHeader[]){
         int CalculatedChecksum;
         int ChecksumFromHeader;
        CalculatedChecksum = ChecksumAMSDOS(pHeader);
        ChecksumFromHeader = (pHeader[67] & 0x0ff) | (pHeader[68] & 0x0ff)<<8;
        if (ChecksumFromHeader == CalculatedChecksum){
            System.out.println("With header");    
            return true;}
            System.out.println("Without header");   
        return false;
}
    
    public void getScores(){
        Object[] oscores = {"Address where scores are located:"};
             String scores = JOptionPane.showInputDialog(dummy, oscores , "88");
             if (scores == null) return;
             try {
             getScores(Util.hexValue(scores));}
             catch (final Exception iox) { }
    }
    
    public void getScores(int address){
            byte value1 = (byte)memory.readByte(address);
            byte value2 = (byte)memory.readByte(address+1);
            String value = Util.hex(value2) + Util.hex(value1);
                    
            try {
            int Value = Util.hexValue(value);
            JOptionPane.showMessageDialog(null, "Score is: "+ Value);}
            
                        catch (final Exception iox) { }
    }
}